/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|          Program Name: Modbus                                                                                         |
|                Author: Michael Krmer                                                                                 |
|                 Email: Michael@dk8pp.de                                                                               |
|                  Date: 21-Jan-2021                                                                                    |
|                  Tool: AVR Studio 7.0                                                                                 |
|                Device: ATTiny, ATMEGA, ATXMEGA                                                                        |
|               Version: 0.1                                                                                            |
|        Program status: still testing                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: Modbus functions                                                                                            |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+

Preprocessor-Macros. Define the following preprocessor symbols to achieve the respective behaviour:
====================

SIMULATION:  If SIMULATION is defined, certain hardware waits and optimizations are turned off.
             SIMULATION must be undefined for final hardware.

DEBUG:       DEBUG will replace inline functions with standard functions. __inline is not supported
             if optimizations are entirely disabled.

ToDo:
=====
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Includes                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#include "project.h"
#include "modbus.h"

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       structs and enums                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Function prototypes                                                                                             |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool crc_error_rx(RXBUFFER *rxbuf) ;
void modbus_read_file_record(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc) ;
void modbus_write_multiple_registers(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc) ;
void modbus_write_single_holding_register(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc) ;
void modbus_write_single_coil(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc) ;
void modbus_read_register(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc) ;
void modbus_read_coils(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc) ;

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Inline functions                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       EEPROM data                                                                                                     |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       global variables                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       PROGMEM constants                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/
/* Table of CRC values for highorder byte */
static const PROGMEM unsigned char auchCRCHi[] = {
        0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
        0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
        0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
        0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
        0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
        0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
        0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
        0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
        0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
        0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
        0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
        0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
        0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40,
        0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
        0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40, 0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41,
        0x00, 0xC1, 0x81, 0x40, 0x01, 0xC0, 0x80, 0x41, 0x01, 0xC0, 0x80, 0x41, 0x00, 0xC1, 0x81, 0x40} ;

/* Table of CRC values for loworder byte */
static const PROGMEM char auchCRCLo[] = {
        0x00, 0xC0, 0xC1, 0x01, 0xC3, 0x03, 0x02, 0xC2, 0xC6, 0x06, 0x07, 0xC7, 0x05, 0xC5, 0xC4, 0x04,
        0xCC, 0x0C, 0x0D, 0xCD, 0x0F, 0xCF, 0xCE, 0x0E, 0x0A, 0xCA, 0xCB, 0x0B, 0xC9, 0x09, 0x08, 0xC8,
        0xD8, 0x18, 0x19, 0xD9, 0x1B, 0xDB, 0xDA, 0x1A, 0x1E, 0xDE, 0xDF, 0x1F, 0xDD, 0x1D, 0x1C, 0xDC,
        0x14, 0xD4, 0xD5, 0x15, 0xD7, 0x17, 0x16, 0xD6, 0xD2, 0x12, 0x13, 0xD3, 0x11, 0xD1, 0xD0, 0x10,
        0xF0, 0x30, 0x31, 0xF1, 0x33, 0xF3, 0xF2, 0x32, 0x36, 0xF6, 0xF7, 0x37, 0xF5, 0x35, 0x34, 0xF4,
        0x3C, 0xFC, 0xFD, 0x3D, 0xFF, 0x3F, 0x3E, 0xFE, 0xFA, 0x3A, 0x3B, 0xFB, 0x39, 0xF9, 0xF8, 0x38,
        0x28, 0xE8, 0xE9, 0x29, 0xEB, 0x2B, 0x2A, 0xEA, 0xEE, 0x2E, 0x2F, 0xEF, 0x2D, 0xED, 0xEC, 0x2C,
        0xE4, 0x24, 0x25, 0xE5, 0x27, 0xE7, 0xE6, 0x26, 0x22, 0xE2, 0xE3, 0x23, 0xE1, 0x21, 0x20, 0xE0,
        0xA0, 0x60, 0x61, 0xA1, 0x63, 0xA3, 0xA2, 0x62, 0x66, 0xA6, 0xA7, 0x67, 0xA5, 0x65, 0x64, 0xA4,
        0x6C, 0xAC, 0xAD, 0x6D, 0xAF, 0x6F, 0x6E, 0xAE, 0xAA, 0x6A, 0x6B, 0xAB, 0x69, 0xA9, 0xA8, 0x68,
        0x78, 0xB8, 0xB9, 0x79, 0xBB, 0x7B, 0x7A, 0xBA, 0xBE, 0x7E, 0x7F, 0xBF, 0x7D, 0xBD, 0xBC, 0x7C,
        0xB4, 0x74, 0x75, 0xB5, 0x77, 0xB7, 0xB6, 0x76, 0x72, 0xB2, 0xB3, 0x73, 0xB1, 0x71, 0x70, 0xB0,
        0x50, 0x90, 0x91, 0x51, 0x93, 0x53, 0x52, 0x92, 0x96, 0x56, 0x57, 0x97, 0x55, 0x95, 0x94, 0x54,
        0x9C, 0x5C, 0x5D, 0x9D, 0x5F, 0x9F, 0x9E, 0x5E, 0x5A, 0x9A, 0x9B, 0x5B, 0x99, 0x59, 0x58, 0x98,
        0x88, 0x48, 0x49, 0x89, 0x4B, 0x8B, 0x8A, 0x4A, 0x4E, 0x8E, 0x8F, 0x4F, 0x8D, 0x4D, 0x4C, 0x8C,
        0x44, 0x84, 0x85, 0x45, 0x87, 0x47, 0x46, 0x86, 0x82, 0x42, 0x43, 0x83, 0x41, 0x81, 0x80, 0x40} ;

char const PROGMEM Modbus_FileName[] = __FILE__ ;       /* Filename */
char const PROGMEM Modbus_TimeStamp[] = __TIMESTAMP__ ; /* Sun Sep 16 01:03:52 1973 */

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       modbus_read_coils                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Read coils                                                                                                            |
|                                                                                                                       |
| Receive buffer holds the received message for this unit. Prepare response in transmit buffer.                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void modbus_read_coils(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc)
{
uint16_t crc ;
uint16_t addr ;
uint16_t count ;
uint16_t startaddr ;
uint16_t endaddr ;
uint16_t i ;
MODBUS_REGISTER *regptr, *regptrlast ;
uint16_t regcnt ;
bool illegal_data_address ;
bool found ;
uint8_t bitcntr ;
uint16_t (*read_function) (uint16_t) ;

   addr = rxbuf->buffer[2]<<8 | rxbuf->buffer[3] ;
   count = rxbuf->buffer[4]<<8 | rxbuf->buffer[5] ;
   i = 0 ;
   if ((count == 0) || (count > ((TXBUFSIZE-5)*8)))
   {
      modbus_error_response(txbuf, 0x03) ; /* exception code 3: Illegal data value */
   }
   else
   {
      illegal_data_address = false ;
      regptr = (MODBUS_REGISTER *) pgm_read_word(&(desc->coil)) ;
      regcnt = pgm_read_word(&(desc->coil_count)) ;
      regptrlast = regptr + regcnt ;
      found = false ;
      while ((!found) && (regptr < regptrlast))
      { /* find (first) entry */
         startaddr = pgm_read_word(&(regptr->addr)) ;
         endaddr = startaddr + pgm_read_word(&(regptr->count)) ;
         found = ((addr >= startaddr) && (addr < endaddr)) ;
         if (!found) regptr++ ;
      }

      if (!found) illegal_data_address = true ;
      else
      { /* found */
         i = 0 ;
         bitcntr = 0 ;
         read_function = (uint16_t (*) (uint16_t)) pgm_read_word(&(regptr->read)) ; /* get pointer to coil read function */
         illegal_data_address |= (read_function == NULL) ;
         illegal_data_address |= ((addr+count) > endaddr) ;
         while ((count > 0) && (!illegal_data_address))
         {
            if (bitcntr == 0) txbuf->buffer[i+3] = 0 ; /* init next byte */
            /* read coil and set response accordingly */
            if (read_function(addr) != 0) txbuf->buffer[i+3] |= (1<<bitcntr) ;
            count-- ;
            if (count > 0)
            {
               addr++ ;
               bitcntr++ ;
               if (bitcntr == 8)
               {
                  i++ ;
                  bitcntr = 0 ;
               }
            }
         }
      }

      if (!illegal_data_address)
      {
         txbuf->buffer[2] = (uint8_t) i+1 ; /* byte count */
         crc = modbus_crc16((uint8_t *) txbuf->buffer, i+4) ;
         txbuf->buffer[i+4] = (uint8_t) (crc) ; /* CRC low byte */
         txbuf->buffer[i+5] = (uint8_t) (crc>>8) ; /* CRC high byte */
         txbuf->count = i+6 ;
         txbuf->completed = false ;
      }
      else modbus_error_response(txbuf, 0x02) ; /* exception code 2: illegal data address */
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       modbus_read_register                                                                                            |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Read single input or holding register                                                                                 |
|                                                                                                                       |
| Receive buffer holds the received message for this unit. Prepare response in transmit buffer.                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void modbus_read_register(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc)
{
uint16_t crc ;
uint16_t addr ;
uint16_t count ;
uint16_t startaddr ;
uint16_t endaddr ;
uint16_t i ;
uint16_t data ;
MODBUS_REGISTER *regptr, *regptrlast ;
uint16_t regcnt ;
uint8_t datatype ;
uint16_t offset ;
bool illegal_data_address ;
bool found ;
uint8_t function_code ;
uint8_t *dataaddr ;
uint16_t (*read_function) (uint16_t) ;

   addr = rxbuf->buffer[2]<<8 | rxbuf->buffer[3] ;
   count = rxbuf->buffer[4]<<8 | rxbuf->buffer[5] ;
   data = count ;
   function_code = rxbuf->buffer[1] ;
   i = 0 ;

   if ((count == 0) || (count > ((TXBUFSIZE-5)/2)))
   {
      modbus_error_response(txbuf, 0x03) ; /* exception code 3: Illegal data value */
   }
   else
   {
      illegal_data_address = false ;
      regptr = NULL ;
      regcnt = 0 ;
      if (function_code == 0x03)
      { /* read holding registers */
         regptr = (MODBUS_REGISTER *) pgm_read_word(&(desc->hreg)) ;
         regcnt = pgm_read_word(&(desc->hreg_count)) ;
      }
      if (function_code == 0x04)
      { /* read input registers */
         regptr = (MODBUS_REGISTER *) pgm_read_word(&(desc->ireg)) ;
         regcnt = pgm_read_word(&(desc->ireg_count)) ;
      }
      regptrlast = regptr + regcnt ;
      found = false ;

      while ((!found) && (regptr < regptrlast))
      { /* find (first) entry */
         startaddr = pgm_read_word(&(regptr->addr)) ;
         endaddr = startaddr + pgm_read_word(&(regptr->count)) ;
         offset = startaddr - pgm_read_word(&(regptr->value)) ;
         found = ((addr >= startaddr) && (addr < endaddr)) ;
         if (!found) regptr++ ;
      }

      if (!found) illegal_data_address = true ;
      else
      { /* found */
         datatype = pgm_read_byte(&(regptr->type)) ;

         if (datatype & T_IMMEDIATE)
         { /* immediate access */
            i = 0 ;
            while ((regptr < regptrlast) && (count > 0) && (!illegal_data_address))
            {
               if ((pgm_read_word(&(regptr->addr)) == addr) && (pgm_read_byte(&(regptr->type))) & T_IMMEDIATE)
               { /* valid address */
                  data = pgm_read_word(&(regptr->value)) ;
                  txbuf->buffer[2*i+3] = (uint8_t) (data>>8) ; /* data 1 */
                  txbuf->buffer[2*i+4] = (uint8_t) (data)    ; /* data 2 */
                  i++ ;
                  count-- ;
                  regptr++ ;
                  addr++ ;
               }
               else illegal_data_address = true ;
            }
         }

         if ((datatype & T_DATAPTR) || (datatype & T_MCBPTR))
         { /* access data memory or MCB */
            if (datatype & T_MCBPTR)
            { /* MCB access, read MCB from EEPROM into data RAM */
               read_function = (uint16_t (*) (uint16_t)) pgm_read_word(&(regptr->read)) ; /* get pointer to MCB read function */
               if ((read_function == NULL) || (!read_function(0))) /* read mcb from EEPROM to RAM */
               { /* invalid MCB in EEPROM, treat as invalid data address */
                  illegal_data_address = true ;
               }
            }
            i = 0 ;
            while ((regptr < regptrlast) && (count > 0) && (!illegal_data_address))
            {
               if (pgm_read_word(&(regptr->addr)) == addr)
               { /* valid address */
                  dataaddr = (uint8_t *) pgm_read_word(&(regptr->value)) ;
                  datatype = pgm_read_byte(&(regptr->type)) ;
                  txbuf->buffer[2*i+4] = (uint8_t) *dataaddr ;    /* low order byte */
                  if (datatype & T_WORD)
                     txbuf->buffer[2*i+3] = (uint8_t) *(dataaddr+1) ; /* high order byte if word access */
                  else txbuf->buffer[2*i+3] = 0 ; /* high order byte is 0 for byte access */
                  i++ ;
                  count-- ;
                  regptr++ ;
                  addr++ ;
               }
               else illegal_data_address = true ;
            }
         }

         if (datatype & T_EEPPTR)
         { /* EEPROM access, always as words */
            i = 0 ;
            while ((count > 0) && (addr < endaddr))
            { /* insert EEPROM data */
               txbuf->buffer[2*i+3] = eeprom_read_byte((const uint8_t *) (2*(addr-offset)+1)) ; /* data 1 */
               txbuf->buffer[2*i+4] = eeprom_read_byte((const uint8_t *) (2*(addr-offset)+0)) ; /* data 2 */
               i++ ;
               count-- ;
               addr++ ;
            }
         }

         if (datatype & T_PROGPTR)
         { /* PROGMEM access, always as words */
            i = 0 ;
            while ((count > 0) && (addr < endaddr))
            { /* insert PROGMEM data */
               txbuf->buffer[2*i+3] = pgm_read_byte((const uint8_t *) (2*(addr-offset)+1)) ; /* data 1 */
               txbuf->buffer[2*i+4] = pgm_read_byte((const uint8_t *) (2*(addr-offset)+0)) ; /* data 2 */
               i++ ;
               count-- ;
               addr++ ;
            }
         }

         if (datatype & T_EXECUTE)
         { /* execute callback function, just one register read is supported */
            read_function = (uint16_t (*) (uint16_t)) pgm_read_word(&(regptr->read)) ; /* get pointer to read function */
            if (read_function == NULL) illegal_data_address = true ;
            else
            {
               data = read_function(addr) ;
               txbuf->buffer[2*i+3] = (uint8_t) (data>>8) ; /* data 1 */
               txbuf->buffer[2*i+4] = (uint8_t) data ; /* data 2 */
               i++ ;
            }
         }
      }

      if (!illegal_data_address)
      {
         txbuf->buffer[2] = (uint8_t) (i<<1) ; /* byte count */
         crc = modbus_crc16((uint8_t *) txbuf->buffer, 2*i+3) ;
         txbuf->buffer[2*i+3] = (uint8_t) (crc) ; /* CRC low byte */
         txbuf->buffer[2*i+4] = (uint8_t) (crc>>8) ; /* CRC high byte */
         txbuf->count = 2*i+5 ;
         txbuf->completed = false ;
      }
      if (illegal_data_address) modbus_error_response(txbuf, 0x02) ; /* exception code 2: illegal data address */
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       modbus_write_single_coil                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Write single coil                                                                                                     |
|                                                                                                                       |
| Receive buffer holds the received message for this unit. Prepare response in transmit buffer.                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void modbus_write_single_coil(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc)
{
uint16_t crc ;
uint16_t addr ;
uint16_t count ;
uint16_t startaddr ;
uint16_t endaddr ;
uint16_t data ;
MODBUS_REGISTER *regptr, *regptrlast ;
uint16_t regcnt ;
bool illegal_data_address ;
bool found ;
uint16_t (*write_function) (uint16_t,uint16_t) ;

   addr = rxbuf->buffer[2]<<8 | rxbuf->buffer[3] ;
   count = rxbuf->buffer[4]<<8 | rxbuf->buffer[5] ;
   data = count ;
   if ((data != 0x0000) && (data != 0xff00))
   {
      modbus_error_response(txbuf, 0x03) ; /* exception code 3: Illegal data value */
   }
   else
   {
      illegal_data_address = false ;
      regptr = (MODBUS_REGISTER *) pgm_read_word(&(desc->coil)) ;
      regcnt = pgm_read_word(&(desc->coil_count)) ;
      regptrlast = regptr + regcnt ;
      found = false ;
      while ((!found) && (regptr < regptrlast))
      { /* find (first) entry */
         startaddr = pgm_read_word(&(regptr->addr)) ;
         endaddr = startaddr + pgm_read_word(&(regptr->count)) ;
         found = ((addr >= startaddr) && (addr < endaddr)) ;
         if (!found) regptr++ ;
      }

      if (!found) illegal_data_address = true ;
      else
      { /* found */
         write_function = (uint16_t (*) (uint16_t,uint16_t)) pgm_read_word(&(regptr->write)) ; /* get pointer to coil write function */
         illegal_data_address |= (write_function == NULL) ;
      }

      if (!illegal_data_address)
      {
         txbuf->buffer[2] = rxbuf->buffer[2] ; /* copy address from received data */
         txbuf->buffer[3] = rxbuf->buffer[3] ; /* copy address from received data */
         data = write_function(addr, data) ;
         txbuf->buffer[4] = (uint8_t) (data>>8) ;
         txbuf->buffer[5] = (uint8_t)  data ;
         crc = modbus_crc16((uint8_t *) txbuf->buffer, 6) ;
         txbuf->buffer[6] = (uint8_t) (crc) ; /* CRC low byte */
         txbuf->buffer[7] = (uint8_t) (crc>>8) ; /* CRC high byte */
         txbuf->count = 8 ;
         txbuf->completed = false ;
      }
      if (illegal_data_address) modbus_error_response(txbuf, 0x02) ; /* exception code 2: illegal data address */
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       modbus_write_single_holding_register                                                                            |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Write single holding register                                                                                         |
|                                                                                                                       |
| Receive buffer holds the received message for this unit. Prepare response in transmit buffer.                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void modbus_write_single_holding_register(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc)
{
//uint16_t crc ;
uint16_t addr ;
uint16_t count ;
uint16_t startaddr ;
uint16_t endaddr ;
uint16_t i ;
uint16_t data ;
MODBUS_REGISTER *regptr, *regptrlast ;
uint16_t regcnt ;
uint8_t datatype ;
bool illegal_data_address ;
bool found ;
uint8_t *dataaddr ;
uint16_t (*read_function) (uint16_t) ;
uint16_t (*write_function) (uint16_t,uint16_t) ;

   addr = rxbuf->buffer[2]<<8 | rxbuf->buffer[3] ;
   count = rxbuf->buffer[4]<<8 | rxbuf->buffer[5] ;
   data = count ;
   i = 0 ;
   illegal_data_address = false ;
   regptr = (MODBUS_REGISTER *) pgm_read_word(&(desc->hreg)) ;
   regcnt = pgm_read_word(&(desc->hreg_count)) ;
   regptrlast = regptr + regcnt ;
   found = false ;
   while ((!found) && (regptr < regptrlast))
   { /* find (first) entry */
      startaddr = pgm_read_word(&(regptr->addr)) ;
      endaddr = startaddr + pgm_read_word(&(regptr->count)) ;
      found = ((addr >= startaddr) && (addr < endaddr)) ;
      if (!found) regptr++ ;
   }

   if (!found) illegal_data_address = true ;
   else
   { /* found */
      datatype = pgm_read_byte(&(regptr->type)) ;

      if (datatype & T_RO) illegal_data_address = true ;
      else
      { /* location is writeable */
         if ((datatype & T_DATAPTR) || (datatype & T_MCBPTR))
         { /* write data memory or MCB */
            if (datatype & T_MCBPTR)
            { /* read MCB from EEPROM into data RAM */
               read_function = (uint16_t (*) (uint16_t)) pgm_read_word(&(regptr->read)) ; /* get pointer to MCB read function */
               if ((read_function == NULL) || (!read_function(0))) /* read mcb from EEPROM to RAM */
               { /* NULL pointer or invalid MCB in EEPROM, treat as invalid data address */
                  illegal_data_address = true ;
               }
            }

            if (!illegal_data_address)
            { /* write data to RAM */
               dataaddr = (uint8_t *) pgm_read_word(&(regptr->value)) ;
               *dataaddr = (uint8_t) data ;
               if (datatype & T_WORD) *(dataaddr+1) = data>>8 ;

               if (datatype & T_MCBPTR)
               { /* Write modified data from RAM back to EEPROM */
                  write_function = (uint16_t (*) (uint16_t, uint16_t)) pgm_read_word(&(regptr->write)) ; /* get pointer to MCB write function */
                  if ((write_function == NULL) || (!write_function(0,0))) /* write mcb from EEPROM to RAM */
                  { /* invalid MCB in EEPROM, treat as invalid data address */
                     illegal_data_address = true ;
                  }
               }

            }
         }

         if (datatype & T_EXECUTE)
         { /* execute callback function */
            write_function = (uint16_t (*) (uint16_t,uint16_t)) pgm_read_word(&(regptr->write)) ; /* get pointer to coil write function */
            if (write_function == NULL) illegal_data_address = true ;
            else illegal_data_address = write_function(addr, data) ;
         }

      }
   }

   if (!illegal_data_address)
   {
   /* Note that the response is the same as the request. */
      for (i=2;i<8;i++) txbuf->buffer[i] = rxbuf->buffer[i] ;
      txbuf->count = 8 ;
      txbuf->completed = false ;
   }
   else modbus_error_response(txbuf, 0x02) ; /* exception code 2: illegal data address */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       modbus_write_multiple_registers                                                                                 |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Write multiple holding registers, destination must be RAM.                                                            |
|                                                                                                                       |
| Receive buffer holds the received message for this unit. Prepare response in transmit buffer.                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void modbus_write_multiple_registers(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc)
{
uint16_t crc ;
uint16_t addr ;
uint16_t count ;
uint16_t startaddr ;
uint16_t endaddr ;
uint16_t i ;
uint16_t data ;
MODBUS_REGISTER *regptr, *regptrlast ;
uint8_t datatype ;
bool illegal_data_address ;
bool found ;
uint8_t *dataaddr ;

   addr = rxbuf->buffer[2]<<8 | rxbuf->buffer[3] ;
   count = rxbuf->buffer[4]<<8 | rxbuf->buffer[5] ;
   illegal_data_address = false ;
   regptr = (MODBUS_REGISTER *) pgm_read_word(&(desc->hreg)) ;
   regptrlast = regptr + pgm_read_word(&(desc->hreg_count)) ;
   found = false ;
   while ((!found) && (regptr < regptrlast))
   { /* find (first) entry */
      startaddr = pgm_read_word(&(regptr->addr)) ;
      endaddr = startaddr + pgm_read_word(&(regptr->count)) ;
      found = ((addr >= startaddr) && (addr < endaddr)) ;
      if (!found) regptr++ ;
   }

   if (!found) illegal_data_address = true ;
   else
   { /* found */
      i = 0 ;
      while ((count > 0) && (!illegal_data_address) && (regptr < regptrlast))
      {
         datatype = pgm_read_byte(&(regptr->type)) ;
         dataaddr = (uint8_t *) pgm_read_word(&(regptr->value)) ;
         if (!(datatype & T_DATAPTR) || (datatype & T_RO)) illegal_data_address = true ;
         else
         { /* location is writeable and destination is RAM */
            /* get next value from receive buffer */
            data = rxbuf->buffer[2*i+7]<<8 | rxbuf->buffer[2*i+8] ;
            /* write data to destination RAM */
            dataaddr = (uint8_t *) pgm_read_word(&(regptr->value)) ;
            *dataaddr = (uint8_t) data ;
            if (datatype & T_WORD) *(dataaddr+1) = data>>8 ;
         }
         count-- ;
         i++ ;
         regptr++ ;
      }
   }

   if (!illegal_data_address)
   {
      txbuf->buffer[2] = rxbuf->buffer[2] ; /* starting address high */
      txbuf->buffer[3] = rxbuf->buffer[3] ; /* starting address low */
      txbuf->buffer[4] = rxbuf->buffer[4] ; /* count high */
      txbuf->buffer[5] = rxbuf->buffer[5] ; /* count low */
      crc = modbus_crc16((uint8_t *) txbuf->buffer, 6) ;
      txbuf->buffer[6] = (uint8_t) (crc) ; /* CRC low byte */
      txbuf->buffer[7] = (uint8_t) (crc>>8) ; /* CRC high byte */
      txbuf->count = 8 ;
      txbuf->completed = false ;
   }
   else modbus_error_response(txbuf, 0x02) ; /* exception code 2: illegal data address */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       modbus_read_file_record                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Read File Record, only file number 1 supported, only one sub-request supported.                                       |
|                                                                                                                       |
| Receive buffer holds the received message for this unit. Prepare response in transmit buffer.                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void modbus_read_file_record(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc)
{
uint16_t crc ;
uint16_t addr ;
uint16_t count ;
uint16_t i ;
MODBUS_REGISTER *regptr, *regptrlast ;
uint16_t regcnt ;
uint8_t datatype ;
bool illegal_data_address ;
bool found ;
uint16_t filenumber ;
uint16_t recordnumber ;
uint16_t bytecount ;
bool eos ; /* end of string */

   illegal_data_address = false ;
   illegal_data_address |= (rxbuf->buffer[2] != 7) ; /* one sub request is 7 bytes long */
   illegal_data_address |= (rxbuf->buffer[3] != 6) ; /* reference type must be 6 */
   filenumber = rxbuf->buffer[4]<<8 | rxbuf->buffer[5] ;
   illegal_data_address |= (filenumber != 1) ; /* only file number 1 is supported */
   recordnumber = rxbuf->buffer[6]<<8 | rxbuf->buffer[7] ;
   count = rxbuf->buffer[8]<<8 | rxbuf->buffer[9] ;

   if ((count == 0) || (count > ((TXBUFSIZE-7)/2)))
   {
      modbus_error_response(txbuf, 0x03) ; /* exception code 3: Illegal data value */
      return ;
   }

   if (illegal_data_address)
   {
      modbus_error_response(txbuf, 0x02) ; /* exception code 2: illegal data address */
      return ;
   }

   regptr = (MODBUS_REGISTER *) pgm_read_word(&(desc->file)) ;
   regcnt = pgm_read_word(&(desc->file_count)) ;
   regptrlast = regptr + regcnt ;
   found = false ;

   while ((!found) && (regptr < regptrlast))
   { /* find entry for this record */
      found = (recordnumber == pgm_read_word(&(regptr->addr))) ; /* read file record number */
      if (!found) regptr++ ;
   }

   if (!found) illegal_data_address = true ;
   else
   { /* found */
      datatype = pgm_read_byte(&(regptr->type)) ;
      bytecount = pgm_read_word(&(regptr->count)) ; /* number of bytes in this message */
      addr = pgm_read_word(&(regptr->value)) ; /* address of record in PROGMEM */

      if (datatype & T_PROGPTR)
      { /* PROGMEM access */
         i = 0 ;
         eos = false ;
         while ((!eos) && (i < 2*count))
         { /* insert PROGMEM data */
            txbuf->buffer[i+5] = pgm_read_byte((const uint8_t *) (addr+i)) ;
            if (bytecount == 0) /* this is a null terminated string */
               eos = (txbuf->buffer[i+5] == 0) ;
            i++ ;
            if (bytecount != 0) /* number of characters specified */
               eos = (i >= bytecount) ;
         }

         while (i < 2*count)
         { /* pad with 0 */
            txbuf->buffer[i+5] = 0 ;
            i++ ;
         }
      }
      else illegal_data_address = true ;
   }

   if (!illegal_data_address)
   {
      txbuf->buffer[2] = (uint8_t) (2*count+2) ; /* data length */
      txbuf->buffer[3] = (uint8_t) (2*count+1) ; /* file response length */
      txbuf->buffer[4] = 6 ; /* reference type is always 6 */
      crc = modbus_crc16((uint8_t *) txbuf->buffer, 2*count+5) ;
      txbuf->buffer[2*count+5] = (uint8_t) (crc) ; /* CRC low byte */
      txbuf->buffer[2*count+6] = (uint8_t) (crc>>8) ; /* CRC high byte */
      txbuf->count = 2*count+7 ;
      txbuf->completed = false ;
   }
   else modbus_error_response(txbuf, 0x02) ; /* exception code 2: illegal data address */
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       crc_error_rx                                                                                                    |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Check received message for CRC error                                                                                  |
|                                                                                                                       |
| Returns true if CRC error encountered                                                                                 |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
bool crc_error_rx(RXBUFFER *rxbuf)
{
uint16_t crc ;
uint16_t len ;
bool retval ;
   len = (uint16_t) rxbuf->next ;
   crc = modbus_crc16((uint8_t *) rxbuf->buffer, len-2) ;
   retval = false ;
   if (rxbuf->buffer[len-2] != (uint8_t) (crc   )) retval = true ; /* check CRC low byte */
   if (rxbuf->buffer[len-1] != (uint8_t) (crc>>8)) retval = true ; /* check CRC high byte */
   return(retval) ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       modbus_message_received                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Prepare the response to a received message                                                                            |
|                                                                                                                       |
| Receive buffer holds the received message. Check CRC and receipient and handle response.                              |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void modbus_message_received(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc)
{
   if (!crc_error_rx(rxbuf))
   { /* request is ignored if RX CRC error */
      total_message_count_l++ ;
      if (total_message_count_l == 0) total_message_count_h++ ;
      broadcast = (rxbuf->buffer[0] == 0) ; /* broadcast message if module address is 0 */
      if (broadcast || (rxbuf->buffer[0] == module_address))
      { /* it's me or a broadcast */
         message_count_l++ ;
         if (message_count_l == 0) message_count_h++ ;

         modbus_message_handler(rxbuf, txbuf, desc) ;

         if ((!listen_only) && (!broadcast))
         { /* no response in broadcast and listen-only mode */
            transmit(txbuf) ; /* send response */
            wait_for_tx(txbuf) ; /* wait for transmit completion */
         }

         if (restart_communications)
         { /* exit listen only mode and restart UART */
            restart_communications = false ;
            listen_only = false ;
         }
      }
   }
   else
   {
      crc_error_count_l++ ;
      if (crc_error_count_l == 0) crc_error_count_h++ ;
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       modbus_message_handler                                                                                          |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Prepare the response to a received message                                                                            |
|                                                                                                                       |
| Receive buffer holds the received message for this unit. Prepare response in transmit buffer.                         |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void modbus_message_handler(RXBUFFER *rxbuf, TXBUFFER *txbuf, MODBUS_DESCRIPTOR *desc)
{
   txbuf->buffer[0] = rxbuf->buffer[0] ; /* copy slave address for response */
   txbuf->buffer[1] = rxbuf->buffer[1] ; /* copy function code for response */

   switch(rxbuf->buffer[1]) /* check function code */
   {
      case 0x01: /* read coils */
         modbus_read_coils(rxbuf, txbuf, desc) ;
      break ;

      case 0x03: /* read holding registers */
      case 0x04: /* read input registers */
         modbus_read_register(rxbuf, txbuf, desc) ;
      break ;

      case 0x05: /* write single coil */
         modbus_write_single_coil(rxbuf, txbuf, desc) ;
      break ;

      case 0x06: /* Write single holding register */
         modbus_write_single_holding_register(rxbuf, txbuf, desc) ;
      break ;

      case 0x10: /* Write multiple holding registers */
         modbus_write_multiple_registers(rxbuf, txbuf, desc) ;
      break ;

      case 0x14: /* Read File Record */
         modbus_read_file_record(rxbuf, txbuf, desc) ;
      break ;

//      case 0x08: /* Diagnostics */
//
///****** All diagnostics functions are untested ********/
//
//         addr = rxbuf->buffer[2]<<8 | rxbuf->buffer[3] ; /* sub-function */
//         data = rxbuf->buffer[4]<<8 | rxbuf->buffer[5] ;
//
//         /* prepare response assuming function succeeds (otherwise the response will be replaced below) */
//         /* Note that the response is (mostly) the same as the request. */
//         for (i=2;i<8;i++) txbuf->buffer[i] = rxbuf->buffer[i] ;
//         txbuf->count = 8 ;
//         txbuf->completed = false ;
//         illegal_data_value = false ;
//         illegal_data_address = false ;
//         return_data = false ;
//         switch(rxbuf->buffer[1]) /* check function code */
//         {
//            case 0: /* loopback, return query data */
//            break ;
//
//            case 1: /* restart communications */
//               restart_communications = true ;
//            break ;
//
//            case 4: /* force listen only mode */
//               listen_only = true ;
//            break ;
//
//            case 10: /* clear counters and diagnostig register */
//               software_reset_count = 0 ;
//               ext_reset_count = 0 ;
//               bo_reset_count = 0 ;
//               wdt_reset_count = 0 ;
//               reset_count = 0 ;
//               total_message_count_l = 0 ;
//               total_message_count_h = 0 ;
//               message_count_l = 0 ;
//               message_count_h = 0 ;
//               crc_error_count_l = 0 ;
//               crc_error_count_h = 0 ;
//               com_error_count_l = 0 ;
//               com_error_count_h = 0 ;
//            break ;
//
//            case 11: /* return bus message count */
//               data = total_message_count_l ;
//               return_data = true ;
//            break ;
//
//            case 12: /* return bus communication error count */
//               data = total_message_count_l ;
//               return_data = true ;
//            break ;
//
//            default:
//               illegal_data_address = true ;
//            break ;
//         }
//
//         if (return_data)
//         { /* insert response data, the rest of the tx-buffer was already assembled above */
//            txbuf->buffer[4] = (uint8_t) (data) ; /* low byte */
//            txbuf->buffer[5] = (uint8_t) (data>>8) ; /* high byte */
//            crc = modbus_crc16((uint8_t *) txbuf->buffer, 6) ;
//            txbuf->buffer[6] = (uint8_t) (crc) ; /* CRC low byte */
//            txbuf->buffer[7] = (uint8_t) (crc>>8) ; /* CRC high byte */
//         }
//
//      break ;
//
//      case 0x11: /* report server ID. Display file name and compile date and time */
//         txbuf->buffer[3] = 0x73 ; /* server ID (unused) */
//         txbuf->buffer[4] = 0xff ; /* run indicator status (on) */
//         i = 5 ;
//         for (j=0;((txbuf->buffer[i]=pgm_read_byte(&FileName[j]))!=0) && (i<(TXBUFSIZE-2)); i++, j++) ;
//         for (j=0;((txbuf->buffer[i]=pgm_read_byte(&CompileDate[j]))!=0) && (i<(TXBUFSIZE-2)); i++, j++) ;
//         for (j=0;((txbuf->buffer[i]=pgm_read_byte(&CompileTime[j]))!=0) && (i<(TXBUFSIZE-2)); i++, j++) ;
//
//         txbuf->buffer[2] = i-3 ; /* byte count */
//         crc = CRC16((uint8_t *) txbuf->buffer, i) ;
//         txbuf->buffer[i+0] = (uint8_t) (crc) ; /* CRC low byte */
//         txbuf->buffer[i+1] = (uint8_t) (crc>>8) ; /* CRC high byte */
//         txbuf->count = i+2 ;
//         txbuf->completed = false ;
//      break ;

      default: /* generate error response */
         modbus_error_response(txbuf, 0x01) ; /* exception code 1: illegal function */
      break ;
   }
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       modbus_error_response                                                                                           |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
| Generate an error response message                                                                                    |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+

    Modbus error responses:
    01: Illegal function
        The function code received in the query is not an allowable action for the server. This may be
        because the function code is only applicable to newer devices, and was not implemented in the
        unit selected. It could also indicate that the server is in the wrong state to process a request of this
        type, for example because it is unconfigured and is being asked to return register values.

    02: Illegal data address
        The data address received in the query is not an allowable address for the server. More
        specifically, the combination of reference number and transfer length is invalid. For a controller with
        100 registers, the PDU addresses the first register as 0, and the last one as 99. If a request is
        submitted with a starting register address of 96 and a quantity of registers of 4, then this request
        will successfully operate (address-wise at least) on registers 96, 97, 98, 99. If a request is
        submitted with a starting register address of 96 and a quantity of registers of 5, then this request
        will fail with Exception Code 0x02 Illegal Data Address since it attempts to operate on registers
        96, 97, 98, 99 and 100, and there is no register with address 100.

    03: Illegal data value
        A value contained in the query data field is not anallowable value for server. This indicates a fault in
        the structure of the remainder of a complex request, such as that the implied length is
        incorrect. It specifically does NOT mean that a data item submitted for storage in a register has a
        value outside the expectation of the application program, since the MODBUS protocol is unaware
        of the significance of any particular value of any particular register.

    04: Server device failure
        An unrecoverable error occurred while the server was attempting to perform the requested action.

    05: Acknowledge
        Specialized use in conjunction with programming commands.
        The server has accepted the request and is processing it, but a long duration of time will be
        required to do so. This response is returned to prevent a timeout error from occurring in the
        client. The client can next issue a Poll Program Complete message to determine if processing is
        completed.

    06: Server device busy
        Specialized use in conjunction with programming commands.
        The server is engaged in processing a longduration program command. The client should
        retransmit the message later when the server is free.

    08: Memory parity error
        Specialized use in conjunction with function codes 20 and 21 and reference type 6, to indicate that
        the extended file area failed to pass a consistency check.
        The server attempted to read record file, but detected a parity error in the memory. The client
        can retry the request, but service may be required on the server device.

    0A: Gateway path unavailable
        Specialized use in conjunction with gateways, indicates that the gateway was unable to allocate
        an internal communication path from the input port to the output port for processing the request.
        Usually means that the gateway is misconfigured or overloaded.

    0B: Gateway target device failed to respond
        Specialized use in conjunction with gateways, indicates that no response was obtained from the
        target device. Usually means that the device is not present on the network.
*/
void modbus_error_response(TXBUFFER *txbuf, uint8_t errorcode)
{
uint16_t crc ;
   txbuf->buffer[1] |= 0x80 ; /* mark reply as error */
   txbuf->buffer[2] = errorcode ; /* insert error code */
   crc = modbus_crc16((uint8_t *) txbuf->buffer, 3) ;
   txbuf->buffer[3] = (uint8_t) (crc) ; /* CRC low byte */
   txbuf->buffer[4] = (uint8_t) (crc>>8) ; /* CRC high byte */
   txbuf->count = 5 ;
   txbuf->completed = false ;
}

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       modbus_crc16                                                                                                    |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|   Generate CRC16                                                                                                      |
|   The function returns the CRC as a unsigned short type                                                               |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
uint16_t modbus_crc16(uint8_t *puchMsg, uint16_t usDataLen)
{
uint8_t uchCRCHi = 0xFF ; /* high byte of CRC initialized */
uint8_t uchCRCLo = 0xFF ; /* low byte of CRC initialized */
unsigned uIndex ; /* will index into CRC lookup table */

   while (usDataLen--) /* pass through message buffer */
   {
      uIndex = uchCRCLo ^ *puchMsg++ ; /* calculate the CRC */
      uchCRCLo = uchCRCHi ^ pgm_read_byte(&auchCRCHi[uIndex]) ;
      uchCRCHi = pgm_read_byte(&auchCRCLo[uIndex]) ;
   }
   return (uchCRCHi << 8 | uchCRCLo) ;
}


