#include <avr/io.h>

.global wdt_off

wdt_off:
                cli                                     ; turn off global interrupts
                wdr                                     ; reset watchdog timer

                in      r16,_SFR_IO_ADDR(MCUSR)
                andi    r16,~(1<<WDRF)
                out     _SFR_IO_ADDR(MCUSR),r16         ; clear WDRF in MCUSR

                lds     r16,WDTCSR
                ori     r16,((1<<WDCE)|(1<<WDE))        ; set WDCE and WDE, keep prescaler setting
                sts     WDTCSR,r16

                andi    r16,~(1<<WDE)                   ; clear WDE, keep others
                sts     WDTCSR,r16                      ; turn off WDT within four instruction cycles
                sei                                     ; global interrupts back on
                ret

.global wdt_on

wdt_on:
                cli                                     ; turn off global interrupts
                wdr                                     ; WDT might be on already
                in      r16,_SFR_IO_ADDR(MCUSR)         ; clear WDRF in MCUSR
                andi    r16,~(1<<WDRF)
                out     _SFR_IO_ADDR(MCUSR),r16

                ldi     r16,((1<<WDCE)|(1<<WDE))        ; set WDCE and WDE
                sts     WDTCSR,r16

                ldi     r16,((0<<WDCE)|(1<<WDE)|(0<<WDP3)|(1<<WDP2)|(0<<WDP1)|(1<<WDP0)) ; 0,5 s timeout
                sts     WDTCSR,r16                      ; turn on WDT within four instruction cycles
                sei                                     ; global interrupts back on
                ret

.global issue_wdt_reset

issue_wdt_reset:
                cli                                     ; turn off global interrupts
                ldi     r16,((1<<WDCE)|(1<<WDE))        ; set WDCE and WDE
                sts     WDTCSR,r16
                ldi     r16,((0<<WDCE)|(1<<WDE)|(0<<WDP3)|(0<<WDP2)|(0<<WDP1)|(0<<WDP0)) ; 16 ms timeout
                sts     WDTCSR,r16                      ; turn on WDT within four instruction cycles
loop:           jmp     loop                            ; WDT reset is issued after 512 cycles (16 ms)

.global set_system_clock_prescaler

set_system_clock_prescaler:
                ldi     r16,(1<<CLKPCE)
                sts     CLKPR,r16                       ; set Clock Prescaler Change Enable
                ldi     r16,(0<<CLKPS3)|(0<<CLKPS2)|(0<<CLKPS1)|(0<<CLKPS0) ; divide clock by 1
                sts     CLKPR,r16                       ; set prescaler
                ret

;.global change_clock
;
;change_clock:
;                LDI     R25,0x80        ;
;                STS     CLKPR,R25       ;enable updating of CLKPR
;                STS     CLKPR,R24       ;write the new value
;                ret                     ;
;
;.global add_two
;
;add_two:
;                ADD R24,R22             ;val_1 = val_1 + val_2
;                CLR     R25             ;return val_1
;                ret                     ;
;
;.global TIMER0_OVF_vect
;
;TIMER0_OVF_vect:
;                sbi _SFR_IO_ADDR(PIND),0        ;toggle PD.0
;                reti                            ;and done




