﻿/*
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|             File Name: project.h                                                                                      |
|                Author: Michael Krämer                                                                                 |
|                 Email: Michael@dk8pp.de                                                                               |
|                  Date: 21.01.2021                                                                                     |
|                  Tool: AVR Studio 7.0                                                                                 |
|                Device: ATTiny, ATMEGA, ATXMEGA                                                                        |
|               Version: 0.1                                                                                            |
|        Program status: still testing                                                                                  |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
|                                                                                                                       |
|  Purpose: Includes project specific include files.                                                                    |
|                                                                                                                       |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#ifndef PROJECT

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Defines                                                                                                         |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#define PROJECT 1

#ifndef _INLINE_
   #ifdef DEBUG
     #define _INLINE_
   #else
     #define _INLINE_ __inline
   #endif
#endif

#ifndef _EXTERN_
   #ifdef DECLARE_VARS
     #define _EXTERN_
     #define _EQ_(val) = val
     #define _EQ4_(v1,v2,v3,v4) = {v1,v2,v3,v4}
     #define _EQ6_(v1,v2,v3,v4,v5,v6) = {v1,v2,v3,v4,v5,v6}
   #else
     #define _EXTERN_ extern
     #define _EQ_(val)
     #define _EQ4_(v1,v2,v3,v4)
     #define _EQ6_(v1,v2,v3,v4,v5,v6)
   #endif
#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Includes                                                                                                        |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#include <stdint.h>
#include <stddef.h>
#include <stdbool.h>
#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/pgmspace.h>
#include <avr/eeprom.h>
#include <avr/cpufunc.h>
#include <util/delay_basic.h>
#include <util/atomic.h>
#include "UART_ATMEGA644.h"

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       structs and enums                                                                                               |
+-----------------------------------------------------------------------------------------------------------------------+
*/

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Function prototypes                                                                                             |
+-----------------------------------------------------------------------------------------------------------------------+
*/
void wait_for_rx(RXBUFFER *rxb) ;
void wait_for_tx(TXBUFFER *txb) ;

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       Inline functions                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/
#ifndef DEBUG
   __inline void TOINT_enable(void)  {TIFR1=(1<<OCF1B); TIMSK1 |=  (1<<OCIE1B) ;} /* enable Output Compare 1B interrupt */
   __inline void TOINT_disable(void) {TIMSK1 &= ~(1<<OCIE1B) ;} /* disable Output Compare 1B interrupt */
#else
   #ifdef DECLARE_VARS
   void TOINT_enable(void)  {TIFR1=(1<<OCF1B); TIMSK1 |=  (1<<OCIE1B) ;} /* enable Output Compare 1B interrupt */
   void TOINT_disable(void) {TIMSK1 &= ~(1<<OCIE1B) ;} /* disable Output Compare 1B interrupt */
   #else
   void TOINT_enable(void) ; /* enable Output Compare 1B interrupt */
   void TOINT_disable(void) ; /* disable Output Compare 1B interrupt */
   #endif
#endif

/*
+-----------------------------------------------------------------------------------------------------------------------+
|       global variables                                                                                                |
+-----------------------------------------------------------------------------------------------------------------------+
*/

#endif
